@@ -23,8 +23,6 @@ Some slight modifications, but at some point
 we should look at maybe pushing this up to Oslo
 """
 
-from __future__ import annotations  # Remove when only supporting python 3.9+
-
 import contextlib
 import errno
 import io
@@ -32,7 +30,7 @@ import math
 import os
 import re
 import tempfile
-from typing import ContextManager, Generator, Optional  # noqa: H301
+from typing import ContextManager, Generator, Optional
 
 import cryptography
 from cursive import exception as cursive_exception
@@ -104,9 +102,9 @@ image_opts = [
                      'an operator has configured glance property protections '
                      'to make some image properties read-only. Cinder will '
                      '*always* filter out image metadata in the namespaces '
-                     '`os_glance` and `img_signature`; this configuration '
-                     'option allows operators to specify *additional* '
-                     'namespaces to be excluded.',
+                     '`os_glance`, `img_signature` and `signature_verified`; '
+                     'this configuration option allows operators to specify '
+                     '*additional* namespaces to be excluded.',
                 default=[]),
 ]
 
@@ -130,7 +128,8 @@ QEMU_IMG_VERSION = None
 
 COMPRESSIBLE_IMAGE_FORMATS = ('qcow2',)
 
-GLANCE_RESERVED_NAMESPACES = ["os_glance", "img_signature"]
+GLANCE_RESERVED_NAMESPACES = ["os_glance", "img_signature",
+                              "signature_verified"]
 
 
 def validate_stores_id(context: context.RequestContext,
@@ -377,18 +376,20 @@ def check_qemu_img_version(minimum_version: str) -> None:
         raise exception.VolumeBackendAPIException(data=_msg)
 
 
-def _convert_image(prefix: tuple,
-                   source: str,
-                   dest: str,
-                   out_format: str,
-                   out_subformat: Optional[str] = None,
-                   src_format: Optional[str] = None,
-                   run_as_root: bool = True,
-                   cipher_spec: Optional[dict] = None,
-                   passphrase_file: Optional[str] = None,
-                   compress: bool = False,
-                   src_passphrase_file: Optional[str] = None,
-                   disable_sparse: bool = False) -> None:
+def _convert_image(
+        prefix: tuple,
+        source: str,
+        dest: str,
+        out_format: str,
+        out_subformat: Optional[str] = None,
+        src_format: Optional[str] = None,
+        run_as_root: bool = True,
+        cipher_spec: Optional[dict] = None,
+        passphrase_file: Optional[str] = None,
+        compress: bool = False,
+        src_passphrase_file: Optional[str] = None,
+        disable_sparse: bool = False,
+        src_img_info: Optional[imageutils.QemuImgInfo] = None) -> None:
     """Convert image to other format.
 
     NOTE: If the qemu-img convert command fails and this function raises an
@@ -407,6 +408,8 @@ def _convert_image(prefix: tuple,
     :param compress: compress w/ qemu-img when possible (best effort)
     :param src_passphrase_file: filename containing source volume's
                                 luks passphrase
+    :param src_img_info: a imageutils.QemuImgInfo object from this image,
+                            or None
     """
 
     # Check whether O_DIRECT is supported and set '-t none' if it is
@@ -463,17 +466,34 @@ def _convert_image(prefix: tuple,
     # some incredible event this is 0 (cirros image?) don't barf
     if duration < 1:
         duration = 1
-    try:
-        image_size = qemu_img_info(source,
-                                   run_as_root=run_as_root).virtual_size
-    except ValueError as e:
+
+    image_info = src_img_info
+    if not image_info:
+        try:
+            image_info = qemu_img_info(source, run_as_root=run_as_root)
+        except Exception:
+            # NOTE: at this point, the image conversion has already
+            # happened, and all that's left is some performance logging.
+            # So ignoring an exception from qemu_img_info here is not a
+            # security risk.  I'm afraid that if we are too strict here
+            # we will cause a regression, given that the converted image
+            # source could be cinder glance_store, Glance, or one of
+            # cinder's own backend drivers.  The nfs driver knows
+            # to pass in a src_img_info object, but others may not.
+            #
+            # We are catching the most general Exception here for a
+            # similar reason: the image conversion has already happened.
+            # If the conversion raised a ProcessExecutionError, we would
+            # never have reached this point.  But a PEE now is meaningless,
+            # so we ignore it.
+            pass
+    if not image_info or image_info.virtual_size is None:
         msg = ("The image was successfully converted, but image size "
-               "is unavailable. src %(src)s, dest %(dest)s. %(error)s")
-        LOG.info(msg, {"src": source,
-                       "dest": dest,
-                       "error": e})
+               "is unavailable. src %(src)s, dest %(dest)s")
+        LOG.info(msg, {"src": source, "dest": dest})
         return
 
+    image_size = image_info.virtual_size
     fsz_mb = image_size / units.Mi
     mbps = (fsz_mb / duration)
     msg = ("Image conversion details: src %(src)s, size %(sz).2f MB, "
@@ -540,7 +560,8 @@ def convert_image(source: str,
                        passphrase_file=passphrase_file,
                        compress=compress,
                        src_passphrase_file=src_passphrase_file,
-                       disable_sparse=disable_sparse)
+                       disable_sparse=disable_sparse,
+                       src_img_info=data)
 
 
 def resize_image(source: str,
@@ -708,10 +729,9 @@ def get_qemu_data(image_id: str,
         if has_meta:
             if not disk_format_raw:
                 raise exception.ImageUnacceptable(
-                    reason=_("qemu-img is not installed and image is of "
-                             "type %s.  Only RAW images can be used if "
-                             "qemu-img is not installed.") %
-                    disk_format_raw,
+                    reason=_("qemu-img is not installed and image is not of "
+                             "type RAW.  Only RAW images can be used if "
+                             "qemu-img is not installed."),
                     image_id=image_id)
         else:
             raise exception.ImageUnacceptable(
@@ -1076,10 +1096,20 @@ def fetch_to_volume_format(context: context.RequestContext,
                       disable_sparse=disable_sparse)
 
 
+@contextlib.contextmanager
+def chown_if_needed(volume_path: str) -> Generator[None, None, None]:
+    if os.name == 'nt' or os.access(volume_path, os.R_OK):
+        yield
+    else:
+        with utils.temporary_chown(volume_path):
+            yield
+
+
 def upload_volume(context: context.RequestContext,
                   image_service: glance.GlanceImageService,
                   image_meta: dict,
                   volume_path: str,
+                  volume_fd = None,
                   volume_format: str = 'raw',
                   run_as_root: bool = True,
                   compress: bool = True,
@@ -1096,14 +1126,13 @@ def upload_volume(context: context.RequestContext,
         if (image_meta['disk_format'] == volume_format):
             LOG.debug("%s was %s, no need to convert to %s",
                       image_id, volume_format, image_meta['disk_format'])
-            if os.name == 'nt' or os.access(volume_path, os.R_OK):
-                with open(volume_path, 'rb') as image_file:
-                    image_service.update(context, image_id, {},
-                                         tpool.Proxy(image_file),
-                                         store_id=store_id,
-                                         base_image_ref=base_image_ref)
+            if volume_fd is not None:
+                image_service.update(context, image_id, {},
+                                     tpool.Proxy(volume_fd),
+                                     store_id=store_id,
+                                     base_image_ref=base_image_ref)
             else:
-                with utils.temporary_chown(volume_path):
+                with chown_if_needed(volume_path):
                     with open(volume_path, 'rb') as image_file:
                         image_service.update(context, image_id, {},
                                              tpool.Proxy(image_file),
